/*
 * ProFTPD - FTP server daemon
 * Copyright (c) 1997, 1998 Public Flood Software
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * PostgreSQL lookup module for ProFTPD (based on mod_unixpw.c@1.20pre3)
 * $Libraries: -lpq$
 * $Id: mod_pgsql.c,v 1.3 1999/08/05 05:11:45 knok Exp $
 * Copyright 1999 NOKUBI Takatsugu <knok@daionet.gr.jp>
 */

#include "conf.h"

#ifdef HAVE_CRYPT_H
#include <crypt.h>
#endif

#include "privs.h"

#include "libpq-fe.h"

#define HASH_TABLE_SIZE		10

typedef struct _idmap {
  struct _idmap *next,*prev;

  int id;			/* uid or gid */
  char *name;			/* user or group name */
} idmap_t;

static xaset_t *uid_table[HASH_TABLE_SIZE];
static xaset_t *gid_table[HASH_TABLE_SIZE];

#define PG_GECOS "PGuser"
#define PG_SHELL "/bin/false"

#define DEFAULT_PGHOST "localhost"
#define DEFAULT_PGPORT "5432"
#define DEFAULT_DBNAME "proftpd"
#define DEFAULT_USERTAB "user"
#define DEFAULT_MAPUID (65534)
#define DEFAULT_MAPGID (65534)
#define DEFAULT_UIDFIELD "uid"
#define DEFAULT_PWDFIELD "passwd"
#define DEFAULT_HOMEFIELD "home"
#define DEFAULT_DBUSER "wwwdata"
#define SQLSTAT_TMPL "select %s,%s,%s from %s;"
#define PG_UID_FNUM (0)
#define PG_PWD_FNUM (1)
#define PG_HOME_FNUM (2)

static PGconn *dbcon= NULL;
static PGresult *dbres = NULL;
static int dbpos = 0;
static char *pghost;
static char *pgport;
static char *dbname;
static char *dbuser;
static char *dbpasswd;
static char *usertab;
static int encrypted = 0;
static int mapuid = -1;
static int mapgid = -1;
static char *uidfield;
static char *pwdfield;
static char *homefield;
static char *shell;

static char *sqlstat;

static void p_setpwent()
{
  log_debug(DEBUG4, "called p_setpwent()");

  if (dbres != NULL) {
    PQclear(dbres);
    dbres = NULL;
  }
  if (dbcon != NULL) {
    PQfinish(dbcon);
    dbcon = NULL;
  }
  dbpos = 0;
  if (pghost == NULL && pgport == NULL && dbname == NULL)
    return;
  dbcon = PQsetdbLogin(pghost, pgport, NULL, NULL, dbname, dbuser, dbpasswd);
  if (dbcon == NULL)
    log_debug(DEBUG0, "database connection failed (dbcon == NULL)");
  if (PQstatus(dbcon) == CONNECTION_BAD)
    log_debug(DEBUG0, "database connection failed (CONNECTION_BAD) pghost=%s pgport=%s dbname=%s Error:% s", pghost, pgport, dbname, PQerrorMessage(dbcon));
  dbres = PQexec(dbcon, sqlstat);
  log_debug(DEBUG4, "sql statement: %s", sqlstat);
}

static void p_endpwent()
{
  if (dbres != NULL) {
    PQclear(dbres);
    dbres = NULL;
  }
  if (dbcon != NULL) {
    PQfinish(dbcon);
    dbcon = NULL;
  }
  dbpos = 0;
}

static struct passwd *p_getpwent()
{
  /*  static passwd dbpwd = {NULL, NULL, 0, 0, NULL, NULL, NULL};*/
  static struct passwd dbpwd = {NULL, NULL, DEFAULT_MAPUID, DEFAULT_MAPUID, 
			 PG_GECOS, NULL, NULL};
  int tmplen;

  log_debug(DEBUG4, "called p_getpwent");

  dbpwd.pw_shell = shell;
  if (! dbpwd.pw_name) free(dbpwd.pw_name);
  if (! dbpwd.pw_passwd) free(dbpwd.pw_passwd);
  if (! dbpwd.pw_dir) free(dbpwd.pw_dir);
  dbpwd.pw_name = NULL;
  dbpwd.pw_passwd = NULL;
  dbpwd.pw_uid = mapuid;
  dbpwd.pw_gid = mapgid;
  dbpwd.pw_dir = NULL;
  
  if (dbres == NULL)
    p_setpwent();
  if (PQstatus(dbcon) != CONNECTION_OK)
    return NULL;

  switch (PQresultStatus(dbres)) {
  case PGRES_COMMAND_OK:
  case PGRES_TUPLES_OK:
    break;
  default:
    return NULL;
  }

  if (PQntuples(dbres) <= dbpos)
    return NULL;

  dbpwd.pw_name = malloc(PQgetlength(dbres, dbpos, PG_UID_FNUM) + 1);
  strcpy(dbpwd.pw_name, PQgetvalue(dbres, dbpos, PG_UID_FNUM));
  dbpwd.pw_passwd = malloc(PQgetlength(dbres, dbpos, PG_PWD_FNUM) + 1);
  strcpy(dbpwd.pw_passwd, PQgetvalue(dbres, dbpos, PG_PWD_FNUM));
  dbpwd.pw_dir = malloc(PQgetlength(dbres, dbpos, PG_HOME_FNUM) + 1);
  strcpy(dbpwd.pw_dir, PQgetvalue(dbres, dbpos, PG_HOME_FNUM));

  log_debug(DEBUG5, "dbpwd: pw_name=%s pw_passwd=%s pw_uid=%d pw_gid=%d pw_dir=%s", dbpwd.pw_name, dbpwd.pw_passwd, dbpwd.pw_uid, dbpwd.pw_gid, dbpwd.pw_dir);

  dbpos ++;

  return &dbpwd;
}

static struct passwd *p_getpwnam(const char *name)
{  
  struct passwd *pw;

  log_debug(DEBUG4, "called p_getpwnam(%s)", name);
    
  p_setpwent();
  while((pw = p_getpwent()) != NULL) {
    log_debug(DEBUG5, "compare %s, %s", name, pw->pw_name);
    if(!strcmp(name,pw->pw_name))
      break;
  }

  if (pw != NULL)
    log_debug(DEBUG5, "p_getpwnam returns: pw_name=%s", pw->pw_name);
 
  return pw;
}

static struct passwd *p_getpwuid(uid_t uid)
{   
  struct passwd *pw;
  
  p_setpwent();
  while((pw = p_getpwent()) != NULL)
    if(pw->pw_uid == uid)
      break;
 
  return pw;
}

static int _compare_id(idmap_t *m1, idmap_t *m2)
{
  if(m1->id < m2->id)
    return -1;
  if(m1->id > m2->id)
    return 1;
  return 0;
}

static idmap_t *_auth_lookup_id(xaset_t **id_table, int id)
{
  int hash = id % HASH_TABLE_SIZE;
  idmap_t *m;

  if(!id_table[hash])
    id_table[hash] = xaset_create(permanent_pool,
                     (XASET_COMPARE)_compare_id);

  for(m = (idmap_t*)id_table[hash]->xas_list; m; m=m->next)
    if(m->id == id)
      break;

  if(!m || m->id != id) {
    /* Isn't in the table */
    m = (idmap_t*)pcalloc(id_table[hash]->mempool,sizeof(idmap_t));
    m->id = id;
    xaset_insert_sort(id_table[hash],(xasetmember_t*)m,FALSE);
  }

  return m;
}

MODRET pg_setpwent(cmd_rec *cmd)
{
  if (get_param_int(TOPLEVEL_CONF, "AuthPGAuthoritative", FALSE) == 0)
    return DECLINED(cmd);

  p_setpwent();

  return HANDLED(cmd);
}

MODRET pg_endpwent(cmd_rec *cmd)
{
  if (get_param_int(TOPLEVEL_CONF, "AuthPGAuthoritative", FALSE) == 0)
    return DECLINED(cmd);

  p_endpwent();

  return HANDLED(cmd);
}

MODRET pg_getpwent(cmd_rec *cmd)
{
  struct passwd *pw;

  if (get_param_int(TOPLEVEL_CONF, "AuthPGAuthoritative", FALSE) == 0)
    return DECLINED(cmd);

  pw = p_getpwent();

  if(pw)
    return mod_create_data(cmd,pw);
  else
    return DECLINED(cmd);
}

MODRET pg_getpwuid(cmd_rec *cmd)
{
  struct passwd *pw;
  uid_t uid;

  if (get_param_int(TOPLEVEL_CONF, "AuthPGAuthoritative", FALSE) == 0)
    return DECLINED(cmd);

  uid = (uid_t)cmd->argv[0];
  pw = p_getpwuid(uid);

  if(pw)
    return mod_create_data(cmd,pw);
  else
    return DECLINED(cmd);
}

MODRET pg_getpwnam(cmd_rec *cmd)
{
  struct passwd *pw;
  const char *name;

  log_debug(DEBUG4, "called pg_getpwnam() argv[0]=%s", cmd->argv[0]);

  if (get_param_int(TOPLEVEL_CONF, "AuthPGAuthoritative", FALSE) == 0)
    return DECLINED(cmd);

  name = cmd->argv[0];
  pw = p_getpwnam(name);

  if(pw)
    return mod_create_data(cmd,pw);
  else
    return DECLINED(cmd);
}

static char *_get_ppg_info(pool *p, const char *u)
{
  struct passwd *pw;
  char *cpw = NULL;

  log_debug(DEBUG4, "called _get_ppg_info(pool, %s)", u);

  pw = p_getpwnam(u);
  if(pw)
    cpw = pstrdup(p,pw->pw_passwd);

  return cpw;
}

/* high-level auth handlers
 */

/* cmd->argv[0] : user name
 * cmd->argv[1] : cleartext password
 */

MODRET pg_auth(cmd_rec *cmd)
{
  time_t now;
  char *cpw;
  time_t lstchg = -1,max = -1,inact = -1,disable = -11;
  const char *name;

  log_debug(DEBUG4, "called pg_auth()");

  if (get_param_int(TOPLEVEL_CONF, "AuthPGAuthoritative", FALSE) == 0)
    return DECLINED(cmd);

  log_debug(DEBUG4, "called pg_auth enabled");

  name = cmd->argv[0];
  time(&now);

  cpw = _get_ppg_info(cmd->tmp_pool,name);

  if(!cpw)
    return DECLINED(cmd);

  if(auth_check(cmd->tmp_pool,cpw,cmd->argv[0],cmd->argv[1]))
    return DECLINED(cmd);

  /*
  if(lstchg > (time_t)0 && max > (time_t)0 &&
     inact > (time_t)0)
    if(now > lstchg + max + inact)
      return DECLINED(cmd);

  if(disable > (time_t)0 && now > disable)
    return DECLINED(cmd);
  */

  log_debug(DEBUG5, "pg_auth succeed.");

  return HANDLED(cmd);
}

/*
 * cmd->argv[0] = hashed password,
 * cmd->argv[1] = user,
 * cmd->argv[2] = cleartext
 */

MODRET pg_check(cmd_rec *cmd)
{
  static char *pw,*cpw;

  log_debug(DEBUG4, "called pg_check()");

  if (get_param_int(TOPLEVEL_CONF, "AuthPGAuthoritative", FALSE) == 0)
    return DECLINED(cmd);

  cpw = cmd->argv[0];
  pw = cmd->argv[2];

  log_debug(DEBUG4, "encrypted=%d", encrypted);

  if (encrypted) {
    if(strcmp(crypt(pw,cpw),cpw) != 0) {
      log_debug(DEBUG5, "pg_check failed.");
      return DECLINED(cmd);
    }
  } else {
    if(strcmp(pw, cpw) != 0) {
      log_debug(DEBUG5, "pg_check failed.");
      return DECLINED(cmd);
    }
  }

  log_debug(DEBUG5, "pg_check succeed.");
  return HANDLED(cmd);
}

MODRET pg_uid_name(cmd_rec *cmd)
{
  idmap_t *m;
  struct passwd *pw;
  uid_t uid;

  if (get_param_int(TOPLEVEL_CONF, "AuthPGAuthoritative", FALSE) == 0)
    return DECLINED(cmd);

  uid = (uid_t)cmd->argv[0];
  m = _auth_lookup_id(uid_table,uid);

  if(!m->name) {
    /* wasn't cached, so perform a lookup */

    pw = p_getpwuid(uid);

    if(pw)
      m->name = pstrdup(permanent_pool,pw->pw_name);
    else {
      char buf[10];

      sprintf(buf,"%lu",(ULONG)uid);
      m->name = pstrdup(permanent_pool,buf);
    }
  }

  return mod_create_data(cmd,m->name);
}

MODRET pg_name_uid(cmd_rec *cmd)
{
  struct passwd *pw;
  const char *name;

  if (get_param_int(TOPLEVEL_CONF, "AuthPGAuthoritative", FALSE) == 0)
    return DECLINED(cmd);

  name = cmd->argv[0];
  
  pw = p_getpwnam(name);

  if(pw)
    return mod_create_data(cmd,(void*)pw->pw_uid);
  return ERROR(cmd);
}

static int pgsql_init()
{
  log_debug(DEBUG4, "called pgsql_init()");

  memset(uid_table,0,sizeof(uid_table));
  memset(gid_table,0,sizeof(gid_table));

  return 0;
}

static int pgsql_child()
{
  int sqllen;

  log_debug(DEBUG4, "called pgsql_child()");

  pghost = get_param_ptr(TOPLEVEL_CONF, "PGhost", FALSE);
  if (pghost == NULL)
    pghost = DEFAULT_PGHOST;
  pgport = get_param_ptr(TOPLEVEL_CONF, "PGport", FALSE);
  if (pgport == NULL) 
    pgport = DEFAULT_PGPORT;
  dbname = get_param_ptr(TOPLEVEL_CONF, "PGdbname", FALSE);
  if (dbname == NULL)
    dbname = DEFAULT_DBNAME;
  dbuser = get_param_ptr(TOPLEVEL_CONF, "PGdbuser", FALSE);
  if (dbuser == NULL)
    dbuser = DEFAULT_DBUSER;
  dbpasswd = get_param_ptr(TOPLEVEL_CONF, "PGdbpasswd", FALSE);
  usertab = get_param_ptr(TOPLEVEL_CONF, "PGusertable", FALSE);
  if (usertab == NULL)
    usertab = DEFAULT_USERTAB;
  encrypted = get_param_int(TOPLEVEL_CONF, "PGencrypted", FALSE);
  if (encrypted < 0)
    encrypted = 1;
  mapuid = get_param_int(TOPLEVEL_CONF, "PGuidMap", FALSE);
  if (mapuid < 0)
    mapuid = DEFAULT_MAPUID;
  mapgid = get_param_int(TOPLEVEL_CONF, "PGgidMap", FALSE);
  if (mapgid < 0)
    mapgid = DEFAULT_MAPGID;
  shell = get_param_int(TOPLEVEL_CONF, "PGshellMap", FALSE);
  if (shell == NULL)
    shell = PG_SHELL;
  uidfield = get_param_ptr(TOPLEVEL_CONF, "PGuidField", FALSE);
  if (uidfield == NULL)
    uidfield = DEFAULT_UIDFIELD;
  pwdfield = get_param_ptr(TOPLEVEL_CONF, "PGpwdField", FALSE);
  if (pwdfield == NULL)
    pwdfield = DEFAULT_PWDFIELD;
  homefield = get_param_ptr(TOPLEVEL_CONF, "PGhomeField", FALSE);
  if (homefield == NULL)
    homefield = DEFAULT_HOMEFIELD;

  sqllen = strlen(SQLSTAT_TMPL) + strlen(uidfield) + strlen(pwdfield)
    + strlen(homefield) + strlen(usertab);
  sqlstat = pcalloc(permanent_pool, sqllen);
  sprintf(sqlstat, SQLSTAT_TMPL, uidfield, pwdfield, homefield, usertab);

  return 0;
}

/* PostgreSQLǤǧڤԤɤ */
MODRET set_pgauthoritative(cmd_rec *cmd) {
  int b;

  log_debug(DEBUG4, "called set_pgauthoritative()");

  CHECK_ARGS(cmd, 1);		/* οΥå */
  CHECK_CONF(cmd, CONF_ROOT);	/* ꤵ줿֤Υå */

  if ((b = get_boolean(cmd, 1)) == -1)
    CONF_ERROR(cmd, "expected boolean argument.");

  add_config_param("AuthPGAuthoritative", 1, (void *) b);

  return HANDLED(cmd);
}

				/* PostgreSQLۥȤ */
MODRET set_pghost(cmd_rec *cmd) {
  log_debug(DEBUG4, "called set_pghost()");

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT);

  add_config_param_str("PGhost", 1, cmd->argv[1]);
  return HANDLED(cmd);
}

MODRET set_pgport(cmd_rec *cmd) {
  int port;

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT);

  add_config_param_str("PGport", 1, cmd->argv[1]);
  return HANDLED(cmd);
}

MODRET set_pgdbname(cmd_rec *cmd) {
  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT);

  add_config_param_str("PGdbname", 1, cmd->argv[1]);
  return HANDLED(cmd);
}

MODRET set_pgdbuser(cmd_rec *cmd) {
  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT);

  add_config_param_str("PGdbuser", 1, cmd->argv[1]);
  return HANDLED(cmd);
}

MODRET set_pgdbpasswd(cmd_rec *cmd) {
  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT);

  add_config_param_str("PGdbpasswd", 1, cmd->argv[1]);
  return HANDLED(cmd);
}

MODRET set_pgusertable(cmd_rec *cmd) {
  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT);

  add_config_param_str("PGusertable", 1, cmd->argv[1]);
  return HANDLED(cmd);
}

MODRET set_pgencrypted(cmd_rec *cmd) {
  int b;

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT);

  if ((b = get_boolean(cmd, 1)) == -1)
    CONF_ERROR(cmd, "expected boolean argument.");

  add_config_param("PGencrypted", 1, (void *) b);

  return HANDLED(cmd);
}

MODRET set_pguidmap(cmd_rec *cmd) {
  int uid;

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT);

  uid = atoi(cmd->argv[1]);
  if (uid < 0 || uid > 65535)
    CONF_ERROR(cmd, "value must be between 0 and 65535");

  add_config_param("PGuidMap", 1, (void *) uid);
  return HANDLED(cmd);
}

MODRET set_pggidmap(cmd_rec *cmd) {
  int gid;

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT);

  gid = atoi(cmd->argv[1]);
  if (gid < 0 || gid > 65535)
    CONF_ERROR(cmd, "value must be between 0 and 65535");

  add_config_param("PGgidMap", 1, (void *) gid);
  return HANDLED(cmd);
}

MODRET set_pgshellmap(cmd_rec *cmd) {
  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT);

  add_config_param_str("PGshellMap", 1, cmd->argv[1]);
  return HANDLED(cmd);
}

MODRET set_pguidfield(cmd_rec *cmd) {
  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT);

  add_config_param_str("PGuidField", 1, cmd->argv[1]);
  return HANDLED(cmd);
}

MODRET set_pgpwdfield(cmd_rec *cmd) {
  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT);

  add_config_param_str("PGpwdField", 1, cmd->argv[1]);
  return HANDLED(cmd);
}

MODRET set_pghomefield(cmd_rec *cmd) {
  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT);

  add_config_param_str("PGhomeField", 1, cmd->argv[1]);
  return HANDLED(cmd);
}

static conftable pgsql_config[] = {
  { "AuthPGAuthoritative",	set_pgauthoritative,		NULL },
  { "PGhost",			set_pghost,			NULL },
  { "PGport",			set_pgport,			NULL },
  { "PGdbname",			set_pgdbname,			NULL },
  { "PGdbuser",			set_pgdbuser,			NULL },
  { "PGdbpasswd",		set_pgdbpasswd,			NULL },
  { "PGusertable",		set_pgusertable,		NULL },
  { "PGencrypted",		set_pgencrypted,		NULL },
  { "PGuidMap",			set_pguidmap,			NULL },
  { "PGgidMap",			set_pggidmap,			NULL },
  { "PGshellMap",		set_pgshellmap,			NULL },
  { "PGuidField",		set_pguidfield,			NULL },
  { "PGpwdField",		set_pgpwdfield,			NULL },
  { "PGhomeField",		set_pghomefield,		NULL },
  { NULL,			NULL,				NULL }
};

static authtable pgsql_auth[] = {
  { 0,  "setpwent",	pg_setpwent	},
  { 0,  "endpwent",	pg_endpwent	},
  { 0,	"getpwent",	pg_getpwent	},
  { 0,  "getpwnam",	pg_getpwnam	},
  { 0,	"getpwuid",	pg_getpwuid	},
  { 0,  "auth",         pg_auth		},
  { 0,  "check",	pg_check	},
  { 0,  "uid_name",	pg_uid_name	},
  { 0,  "name_uid",	pg_name_uid	},
  { 0,  NULL }
};

module pgsql_module = {
  NULL,NULL,			/* Always NULL */
  0x20,				/* API Version 2.0 */
  "pgsql",
  pgsql_config,		/* Configuration directive table */
  NULL,				/* No command handlers */
  pgsql_auth,			/* Authentication handlers */
  pgsql_init,pgsql_child	/* Initialization functions */
};
